/*
 * Copyright (c) 2000 - 2017 Samsung Electronics Co., Ltd. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

#include <stdlib.h>
#include <stdarg.h>
#include <gio/gio.h>
#include <unistd.h>

#include <vconf.h>

#include "badge.h"
#include "badge_log.h"
#include "badge_error.h"
#include "badge_internal.h"
#include "badge_ipc.h"
#include "badge_setting.h"

#define PROVIDER_BUS_NAME "org.tizen.data_provider_service"
#define PROVIDER_OBJECT_PATH "/org/tizen/data_provider_service"
#define PROVIDER_BADGE_INTERFACE_NAME "org.tizen.data_provider_badge_service"

#define DBUS_SERVICE_DBUS "org.freedesktop.DBus"
#define DBUS_PATH_DBUS "/org/freedesktop/DBus"
#define DBUS_INTERFACE_DBUS "org.freedesktop.DBus"

typedef struct _task_list task_list;
struct _task_list {
	task_list *prev;
	task_list *next;

	void (*task_cb) (void *data);
	void *data;
};
static task_list *g_task_list;

static GDBusConnection *_gdbus_conn = NULL;
static int monitor_id = 0;
static int provider_monitor_id = 0;
static int is_master_started = 0;

static void _do_deferred_task(void);

int badge_ipc_is_master_ready(void)
{
	GVariant *result = NULL;
	GError *err = NULL;
	gboolean name_exist;

	result = g_dbus_connection_call_sync(
			_gdbus_conn,
			DBUS_SERVICE_DBUS,
			DBUS_PATH_DBUS,
			DBUS_INTERFACE_DBUS,
			"NameHasOwner",
			g_variant_new("(s)", PROVIDER_BUS_NAME),
			G_VARIANT_TYPE("(b)"),
			G_DBUS_CALL_FLAGS_NONE,
			-1,
			NULL,
			&err);

	if (err || (result == NULL)) {
		/* LCOV_EXCL_START */
		if (err) {
			ERR("No reply. error = %s", err->message);
			g_error_free(err);
		}
		is_master_started = 0;
		/* LCOV_EXCL_STOP */
	} else {
		g_variant_get(result, "(b)", &name_exist);

		if (!name_exist) {
			/* LCOV_EXCL_START */
			ERR("Name not exist %s", PROVIDER_BUS_NAME);
			ERR("the master has been stopped");
			is_master_started = 0;
			/* LCOV_EXCL_STOP */
		} else {
			DBG("the master has been started");
			is_master_started = 1;
		}
	}

	if (result)
		g_variant_unref(result);

	return is_master_started;
}

int badge_ipc_add_deferred_task(
		void (*badge_add_deferred_task)(void *data),
		void *user_data)
{
	task_list *list;
	task_list *list_new;

	list_new = (task_list *) malloc(sizeof(task_list));

	if (list_new == NULL)
		return BADGE_ERROR_OUT_OF_MEMORY;

	list_new->next = NULL;
	list_new->prev = NULL;

	list_new->task_cb = badge_add_deferred_task;
	list_new->data = user_data;

	if (g_task_list == NULL) {
		g_task_list = list_new;
	} else {
		list = g_task_list;

		while (list->next != NULL)
			list = list->next;

		list->next = list_new;
		list_new->prev = list;
	}
	return BADGE_ERROR_NONE;
}

int badge_ipc_del_deferred_task(
		void (*badge_add_deferred_task)(void *data))
{
	task_list *list_del;
	task_list *list_prev;
	task_list *list_next;

	list_del = g_task_list;

	if (list_del == NULL)
		return BADGE_ERROR_INVALID_PARAMETER;

	while (list_del->prev != NULL)
		list_del = list_del->prev;

	do {
		if (list_del->task_cb == badge_add_deferred_task) {
			list_prev = list_del->prev;
			list_next = list_del->next;

			if (list_prev == NULL)
				g_task_list = list_next;
			else
				list_prev->next = list_next;

			if (list_next == NULL) {
				if (list_prev != NULL)
					list_prev->next = NULL;

			} else {
				list_next->prev = list_prev;
			}

			free(list_del);
			return BADGE_ERROR_NONE;
		}
		list_del = list_del->next;
	} while (list_del != NULL);

	return BADGE_ERROR_INVALID_PARAMETER;
}

/* LCOV_EXCL_START */
static void _do_deferred_task(void)
{
	task_list *list_do;
	task_list *list_temp;

	if (g_task_list == NULL)
		return;

	list_do = g_task_list;
	g_task_list = NULL;

	while (list_do->prev != NULL)
		list_do = list_do->prev;

	while (list_do != NULL) {
		if (list_do->task_cb != NULL) {
			list_do->task_cb(list_do->data);
			DBG("called:%p", list_do->task_cb);
		}
		list_temp = list_do->next;
		free(list_do);
		list_do = list_temp;
	}
}
/* LCOV_EXCL_STOP */

/*
 * dbus handler implementation
 */
/* LCOV_EXCL_START */
static void _insert_badge_notify(GVariant *parameters)
{
	char *pkgname = NULL;
	uid_t uid;

	g_variant_get(parameters, "(&si)", &pkgname, &uid);
	badge_changed_cb_call(BADGE_ACTION_CREATE, pkgname, 0, uid);
}
/* LCOV_EXCL_STOP */

/* LCOV_EXCL_START */
static void _delete_badge_notify(GVariant *parameters)
{
	char *pkgname = NULL;
	uid_t uid;

	g_variant_get(parameters, "(&si)", &pkgname, &uid);
	badge_changed_cb_call(BADGE_ACTION_REMOVE, pkgname, 0, uid);
}
/* LCOV_EXCL_STOP */

/* LCOV_EXCL_START */
static void _set_badge_count_notify(GVariant *parameters)
{
	char *pkgname = NULL;
	int count = 0;
	uid_t uid;

	g_variant_get(parameters, "(&sii)", &pkgname, &count, &uid);
	badge_changed_cb_call(BADGE_ACTION_UPDATE, pkgname, count, uid);
}
/* LCOV_EXCL_STOP */

/* LCOV_EXCL_START */
static void _set_disp_option_notify(GVariant *parameters)
{
	char *pkgname = NULL;
	int is_display = 0;
	uid_t uid;

	g_variant_get(parameters, "(&sii)", &pkgname, &is_display, &uid);
	badge_changed_cb_call(BADGE_ACTION_CHANGED_DISPLAY, pkgname, is_display, uid);
}
/* LCOV_EXCL_STOP */

/* LCOV_EXCL_START */
static void _handle_badge_notify(GDBusConnection *connection,
		const gchar     *sender_name,
		const gchar     *object_path,
		const gchar     *interface_name,
		const gchar     *signal_name,
		GVariant        *parameters,
		gpointer         user_data)
{
	DBG("signal_name: %s", signal_name);
	if (g_strcmp0(signal_name, "insert_badge_notify") == 0)
		_insert_badge_notify(parameters);
	else if (g_strcmp0(signal_name, "delete_badge_notify") == 0)
		_delete_badge_notify(parameters);
	else if (g_strcmp0(signal_name, "set_badge_count_notify") == 0)
		_set_badge_count_notify(parameters);
	else if (g_strcmp0(signal_name, "set_disp_option_notify") == 0)
		_set_disp_option_notify(parameters);
}
/* LCOV_EXCL_STOP */

static int _dbus_init(void)
{
	GError *error = NULL;

	if (_gdbus_conn == NULL) {
		_gdbus_conn = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, &error);
		if (_gdbus_conn == NULL) {
			/* LCOV_EXCL_START */
			if (error != NULL) {
				ERR("Failed to get dbus [%s]", error->message);
				g_error_free(error);
			}
			/* LCOV_EXCL_STOP */
			return BADGE_ERROR_IO_ERROR;
		}
		badge_error_quark();
	}

	return BADGE_ERROR_NONE;
}

static int _dbus_signal_init()
{
	int ret = BADGE_ERROR_NONE;
	int id;

	if (monitor_id == 0) {
		DBG("get dbus connection success");
		id = g_dbus_connection_signal_subscribe(_gdbus_conn,
				PROVIDER_BUS_NAME,
				PROVIDER_BADGE_INTERFACE_NAME,	/*   interface */
				NULL,				/*   member */
				PROVIDER_OBJECT_PATH,		/*   path */
				NULL,				/*   arg0 */
				G_DBUS_SIGNAL_FLAGS_NONE,
				_handle_badge_notify,
				NULL,
				NULL);

		DBG("subscribe id : %d", id);
		if (id == 0) {
			/* LCOV_EXCL_START */
			ret = BADGE_ERROR_IO_ERROR;
			ERR("Failed to _register_noti_dbus_interface");
			/* LCOV_EXCL_STOP */
		} else {
			monitor_id = id;
		}
	}

	return ret;
}

/*
 * implement user request
 */
static int _send_sync_badge(GVariant *body, GDBusMessage **reply, char *cmd)
{
	GError *err = NULL;
	GDBusMessage *msg;
	int ret = BADGE_ERROR_NONE;

	msg = g_dbus_message_new_method_call(
			PROVIDER_BUS_NAME,
			PROVIDER_OBJECT_PATH,
			PROVIDER_BADGE_INTERFACE_NAME,
			cmd);
	if (!msg) {
		/* LCOV_EXCL_START */
		ERR("Can't allocate new method call");
		if (body)
			g_variant_unref(body);
		return BADGE_ERROR_OUT_OF_MEMORY;
		/* LCOV_EXCL_STOP */
	}

	if (body != NULL)
		g_dbus_message_set_body(msg, body);

	*reply = g_dbus_connection_send_message_with_reply_sync(
			_gdbus_conn,
			msg,
			G_DBUS_SEND_MESSAGE_FLAGS_NONE,
			-1,
			NULL,
			NULL,
			&err);

	g_object_unref(msg);

	if (!*reply) {
		/* LCOV_EXCL_START */
		ret = BADGE_ERROR_SERVICE_NOT_READY;
		if (err != NULL) {
			ERR("No reply. cmd = %s,  error = %s", cmd, err->message);
			if (err->code == G_DBUS_ERROR_ACCESS_DENIED)
				ret = BADGE_ERROR_PERMISSION_DENIED;
			g_error_free(err);
		}
		return ret;
		/* LCOV_EXCL_STOP */
	}

	if (g_dbus_message_to_gerror(*reply, &err)) {
		ret = err->code;
		ERR("_send_sync_badge cmd = %s, error %s", cmd, err->message);
		if (err->code == G_DBUS_ERROR_ACCESS_DENIED)
			ret = BADGE_ERROR_PERMISSION_DENIED;
		g_error_free(err);
		return ret;
	}
	DBG("_send_sync_badge done !!");
	return BADGE_ERROR_NONE;
}

static int _send_service_register(uid_t uid)
{
	GDBusMessage *reply = NULL;
	int result;

	result = _send_sync_badge(g_variant_new("(i)", uid), &reply, "badge_service_register");

	if (reply)
		g_object_unref(reply);

	DBG("_send_service_register dones");
	return result;
}

static int _ipc_monitor_register(uid_t uid)
{
	DBG("register a service\n");

	return  _send_service_register(uid);
}

/* LCOV_EXCL_START */
static void _on_name_appeared(GDBusConnection *connection,
		const gchar     *name,
		const gchar     *name_owner,
		gpointer         user_data)
{
	DBG("name appeared : %s", name);
	is_master_started = 1;
	_ipc_monitor_register(GPOINTER_TO_INT(user_data));

	_do_deferred_task();
}
/* LCOV_EXCL_STOP */

/* LCOV_EXCL_START */
static void _on_name_vanished(GDBusConnection *connection,
		const gchar     *name,
		gpointer         user_data)
{
	DBG("name vanished : %s", name);
	is_master_started = 0;
}
/* LCOV_EXCL_STOP */

int badge_ipc_monitor_init(uid_t uid)
{
	DBG("register a service\n");
	int ret = BADGE_ERROR_NONE;

	ret = _dbus_init();
	if (ret != BADGE_ERROR_NONE) {
		/* LCOV_EXCL_START */
		ERR("Can't init dbus %d", ret);
		return ret;
		/* LCOV_EXCL_STOP */
	}

	ret = _dbus_signal_init();
	if (ret != BADGE_ERROR_NONE) {
		/* LCOV_EXCL_START */
		ERR("Can't init dbus signal %d", ret);
		return ret;
		/* LCOV_EXCL_STOP */
	}

	ret = _ipc_monitor_register(uid);
	if (ret != BADGE_ERROR_NONE) {
		/* LCOV_EXCL_START */
		ERR("Can't init ipc_monitor_register %d", ret);
		return ret;
		/* LCOV_EXCL_STOP */
	}

	if (provider_monitor_id == 0) {
		provider_monitor_id = g_bus_watch_name_on_connection(
				_gdbus_conn,
				PROVIDER_BUS_NAME,
				G_BUS_NAME_WATCHER_FLAGS_NONE,
				_on_name_appeared,
				_on_name_vanished,
				GINT_TO_POINTER((int)uid),
				NULL);

		if (provider_monitor_id == 0) {
			/* LCOV_EXCL_START */
			ERR("watch on name fail");
			g_dbus_connection_signal_unsubscribe(_gdbus_conn, monitor_id);
			monitor_id = 0;
			return BADGE_ERROR_IO_ERROR;
			/* LCOV_EXCL_STOP */
		}
	}

	return ret;
}

int badge_ipc_monitor_fini(void)
{
	if (provider_monitor_id) {
		g_bus_unwatch_name(provider_monitor_id);
		provider_monitor_id = 0;
	}

	if (monitor_id) {
		g_dbus_connection_signal_unsubscribe(_gdbus_conn, monitor_id);
		monitor_id = 0;
	}
	return BADGE_ERROR_NONE;
}

int badge_ipc_request_is_existing(const char *pkgname, bool *existing, uid_t uid)
{
	int result;
	GDBusMessage *reply = NULL;
	GVariant *body;
	GVariant *reply_body;
	int ret_existing;

	result = _dbus_init();
	if (result != BADGE_ERROR_NONE) {
		ERR("Can't init dbus %d", result);
		return result;
	}
	body = g_variant_new("(si)", pkgname, uid);

	result = _send_sync_badge(body, &reply, "get_badge_existing");
	if (result == BADGE_ERROR_NONE) {
		reply_body = g_dbus_message_get_body(reply);
		g_variant_get(reply_body, "(i)", &ret_existing);
		*existing = ret_existing;
	}

	if (reply)
		g_object_unref(reply);

	DBG("badge_ipc_request_is_existing done [result: %d]", result);
	return result;
}

int badge_ipc_request_get_list(badge_foreach_cb callback, void *data, uid_t uid)
{
	GDBusMessage *reply = NULL;
	int result;
	GVariant *reply_body;
	GVariant *iter_body;
	GVariant *body;
	GVariantIter *iter;
	badge_info_s badge;

	result = _dbus_init();
	if (result != BADGE_ERROR_NONE) {
		ERR("Can't init dbus %d", result);
		return result;
	}

	body = g_variant_new("(i)", uid);
	if (!body) {
		ERR("Cannot create gvariant. Out of memory.");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	result = _send_sync_badge(body, &reply, "get_list");
	if (result == BADGE_ERROR_NONE) {
		reply_body = g_dbus_message_get_body(reply);
		g_variant_get(reply_body, "(a(v))", &iter);
		while (g_variant_iter_loop(iter, "(v)", &iter_body)) {
			g_variant_get(iter_body, "(&si)", &badge.pkg, &badge.badge_count);
			DBG("call calback : %s", badge.pkg);
			callback(badge.pkg, badge.badge_count, data);
		}
		g_variant_iter_free(iter);
	}

	if (reply)
		g_object_unref(reply);

	return result;
}

int badge_ipc_request_insert(const char *pkgname, const char *writable_pkg, const char *caller, uid_t uid)
{
	int result;
	GDBusMessage *reply = NULL;
	GVariant *body;

	result = _dbus_init();
	if (result != BADGE_ERROR_NONE) {
		ERR("Can't init dbus %d", result);
		return result;
	}

	body = g_variant_new("(sssi)", pkgname, writable_pkg, caller, uid);
	if (!body) {
		ERR("Cannot create gvariant. Out of memory.");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	result = _send_sync_badge(body, &reply, "insert_badge");

	if (reply)
		g_object_unref(reply);

	DBG("badge_ipc_request_insert done [result: %d]", result);
	return result;
}

int badge_ipc_request_delete(const char *pkgname, const char *caller, uid_t uid)
{
	int result;
	GDBusMessage *reply = NULL;
	GVariant *body;

	result = _dbus_init();
	if (result != BADGE_ERROR_NONE) {
		ERR("Can't init dbus %d", result);
		return result;
	}

	body = g_variant_new("(ssi)", pkgname, caller, uid);
	if (!body) {
		ERR("Cannot create gvariant. Out of memory.");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	result = _send_sync_badge(body, &reply, "delete_badge");

	if (reply)
		g_object_unref(reply);

	DBG("badge_ipc_request_delete done [result: %d]", result);
	return result;
}

int badge_ipc_request_set_count(const char *pkgname, const char *caller, int count, uid_t uid)
{
	int result;
	GDBusMessage *reply = NULL;
	GVariant *body;

	result = _dbus_init();
	if (result != BADGE_ERROR_NONE) {
		ERR("Can't init dbus %d", result);
		return result;
	}

	body = g_variant_new("(ssii)", pkgname, caller, count, uid);
	if (!body) {
		ERR("Cannot create gvariant. Out of memory.");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	result = _send_sync_badge(body, &reply, "set_badge_count");

	if (reply)
	    g_object_unref(reply);

	DBG("badge_ipc_request_set_count done [result: %d]", result);
	return result;
}

int badge_ipc_request_get_count(const char *pkgname, unsigned int *count, uid_t uid)
{
	int result;
	GDBusMessage *reply = NULL;
	GVariant *body;
	GVariant *reply_body;
	int ret_count;

	result = _dbus_init();
	if (result != BADGE_ERROR_NONE) {
		ERR("Can't init dbus %d", result);
		return result;
	}

	body = g_variant_new("(si)", pkgname, uid);
	if (!body) {
		ERR("Cannot create gvariant. Out of memory.");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	result = _send_sync_badge(body, &reply, "get_badge_count");
	if (result == BADGE_ERROR_NONE) {
		reply_body = g_dbus_message_get_body(reply);
		g_variant_get(reply_body, "(i)", &ret_count);
		*count = ret_count;
	}

	if (reply)
		g_object_unref(reply);

	DBG("badge_ipc_request_get_count done [result: %d]", result);
	return result;
}

int badge_ipc_request_set_display(const char *pkgname, const char *caller,
				  unsigned int display_option, uid_t uid)
{
	int result;
	GDBusMessage *reply = NULL;
	GVariant *body;

	result = _dbus_init();
	if (result != BADGE_ERROR_NONE) {
		ERR("Can't init dbus %d", result);
		return result;
	}
	body = g_variant_new("(ssii)", pkgname, caller, display_option, uid);
	if (!body) {
		ERR("Cannot create gvariant. Out of memory.");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	result = _send_sync_badge(body, &reply, "set_disp_option");
	if (reply)
		g_object_unref(reply);

	DBG("badge_ipc_request_set_display done [result: %d]", result);
	return result;
}

int badge_ipc_request_get_display(const char *pkgname, unsigned int *is_display, uid_t uid)
{
	int result ;
	GDBusMessage *reply = NULL;
	GVariant *body ;
	GVariant *reply_body;
	unsigned int ret_is_display;

	result = _dbus_init();
	if (result != BADGE_ERROR_NONE) {
		ERR("Can't init dbus %d", result);
		return result;
	}

	body = g_variant_new("(si)", pkgname, uid);
	if (!body) {
		ERR("Cannot create gvariant. Out of memory.");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	result = _send_sync_badge(body, &reply, "get_disp_option");
	if (result == BADGE_ERROR_NONE) {
		reply_body = g_dbus_message_get_body(reply);
		g_variant_get(reply_body, "(i)", &ret_is_display);
		*is_display = ret_is_display;
	}

	if (reply)
		g_object_unref(reply);

	DBG("badge_ipc_request_get_display done [result: %d]", result);
	return result;
}

int badge_ipc_request_update_setting(badge_setting_h setting, uid_t uid)
{
	int ret;
	GDBusMessage *reply = NULL;
	GVariant *body = NULL;

	ret = _dbus_init();
	if (ret != BADGE_ERROR_NONE) {
		ERR("Can't init dbus %d", ret);
		return ret;
	}

	body = g_variant_new("(ssii)",
			     setting->pkgname,
			     setting->appid,
			     (int)(setting->allow_to_display),
			     uid);

	ret = _send_sync_badge(body, &reply, "update_badge_setting");
	if (ret != BADGE_ERROR_NONE)
		ERR("Failed badge update setting");

	if (reply)
		g_object_unref(reply);

	return ret;
}

int badge_ipc_request_get_setting_by_appid(badge_setting_h *setting, const char *appid, uid_t uid)
{
	int ret;
	GDBusMessage *reply = NULL;
	GVariant *body = NULL;
	GVariant *reply_body = NULL;
	GVariant *setting_body = NULL;
	badge_setting_h result_setting;

	ret = _dbus_init();
	if (ret != BADGE_ERROR_NONE) {
		ERR("Can't init dbus %d", ret);
		return ret;
	}

	body = g_variant_new("(si)", appid, uid);

	ret = _send_sync_badge(body, &reply, "get_setting_by_appid");
	if (ret == BADGE_ERROR_NONE) {
		reply_body = g_dbus_message_get_body(reply);
		g_variant_get(reply_body, "(v)", &setting_body);

		result_setting = (struct badge_setting *)malloc(sizeof(struct badge_setting));
		if (result_setting == NULL) {
			ERR("Failed memory allocation.");
			g_object_unref(reply);
			return BADGE_ERROR_OUT_OF_MEMORY;
		}
		badge_ipc_make_setting_from_gvariant(result_setting, setting_body);
		*setting = result_setting;
		g_variant_unref(setting_body);
	}

	if (reply)
		g_object_ref(reply);

	return ret;
}

EXPORT_API int badge_ipc_make_setting_from_gvariant(badge_setting_h setting, GVariant *variant)
{
	char *pkgname;
	char *appid;
	int allow_to_display;

	if (setting == NULL || variant == NULL) {
		DBG("Invalid Parameter");
		return BADGE_ERROR_INVALID_PARAMETER;
	}

	g_variant_get(variant, "(&s&si)", &pkgname, &appid, &allow_to_display);

	setting->pkgname = strdup(pkgname);
	setting->appid = strdup(appid);
	setting->allow_to_display = allow_to_display;

	return BADGE_ERROR_NONE;
}

EXPORT_API GVariant *badge_ipc_make_gvariant_from_setting(badge_setting_h setting)
{
	GVariant *body = NULL;

	body = g_variant_new("(ssi)",
			     setting->pkgname,
			     setting->appid,
			     setting->allow_to_display);

	return body;
}

int badge_ipc_init_badge(uid_t uid)
{
	int ret;
	GDBusMessage *reply = NULL;
	GVariant *body;

	body = g_variant_new("(i)", uid);
	if (!body) {
		ERR("Cannot create gvariant.");
		return BADGE_ERROR_OUT_OF_MEMORY;
	}

	ret = _send_sync_badge(body, &reply, "init_badge");

	if (reply)
		g_object_unref(reply);

	DBG("badge_ipc_init_badge done[result:%d]", ret);
	return ret;
}
